import math
import RemoteApi_pb2
import remoteapi

"""
TestRfProp.py

Unit test for Remote API that touches all the RF Prop API methods.
"""

ReturnCodeOnly = lambda r: r.response.returnCode

client = remoteapi.DiscoveryClient().createClient()
if client is None:
	raise Exception("Failed to connect.")

pi = client.pluginApi()
piData = pi.data()
piSimdis = pi.simdis()

f = piData.deleteScenario()
# Deleting a scenario in Debug mode can take a while to respond...
if not f.isOk(10000):
	raise Exception("Failed deleteScenario")

# Load data file is unique because it can take a while to respond...
f = piData.loadDataFile("$(SIMDIS_DIR)/demos/RFProp/AREPS/RCSnAREPS.zip")
print("Requested load of data file, waiting up to 50 seconds")
if not f.isOk(50000):
	raise Exception("Failed loadDataFile: " + f.responseDescription())

# Ensure the scenario got loaded properly with a find-beam
f = piData.findBeam('Search Radar')
if not f.isOk(2000) or f.value() == 0:
	raise Exception("Failed findBeam on Search Radar")
searchRadarId = f.value()

# Set the data time to 2 minutes into the scenario (024 2003 19:20:33.00)
piData.clockSetTime(2056833.0)

############################################
# Test GetPOD
pod = piSimdis.getPOD(searchRadarId, math.radians(91), 12000, 100)

if not pod.isOk(2000):
	raise Exception("Failed to get a response rfpropGetPOD")
if not 99.5 <= pod.value() <= 100.0:
	raise Exception('Got a Pod outside range.  Expected 99.9, got ' + str(pod.value()))

############################################
# Test GetLoss
loss = piSimdis.getLoss(searchRadarId, math.radians(91), 12000, 100)
if not loss.isOk(2000):
	raise Exception("Failed to get a response rfpropGetLoss")
if not 121.0 <= loss.value() <= 122.0:
	raise Exception('Got a Loss outside range.  Expected 121.4, got ' + str(loss.value()))

############################################
# Test GetPPF
ppf = piSimdis.getPPF(searchRadarId, math.radians(91), 12000, 100)
if not ppf.isOk(2000):
	raise Exception("Failed to get a response rfpropGetPPF")
if not 2.0 <= ppf.value() <= 2.2:
	raise Exception('Got a PPF outside range.  Expected 2.1, got ' + str(ppf.value()))

############################################
# Test GetSNR
snr = piSimdis.getSNR(searchRadarId, math.radians(91), 3400, 120, 82.3, 60, 100, 1300)
if not snr.isOk(2000):
	raise Exception("Failed to get a response rfpropGetSNR")
if not 59.2 <= snr.value() <= 59.4:
	raise Exception('Got SNR outside range.  Expected 59.29, got ' + str(snr.value()))

############################################
# Test GetCNR
cnr = piSimdis.getCNR(searchRadarId, math.radians(91), 12000)

if not cnr.isOk(2000):
	raise Exception("Failed to get a response rfpropGetCNR")
if not 19.1 <= cnr.value() <= 20.1:
	raise Exception('Got CNR outside range.  Expected 20.0, got ' + str(cnr.value()))

############################################
# Test GetOneWayPower
oneWayPower = piSimdis.getOneWayPower(searchRadarId, math.radians(91), 12000,3400, 82.3, 1200, 60)

if not oneWayPower.isOk(2000):
	raise Exception("Failed to get a response rfpropGetOneWayPower")
if not 17.4 <= oneWayPower.value() <= 17.6:
	raise Exception('Got One Way Power Received outside range.  Expected 17.53, got ' + str(oneWayPower.value()))

############################################
# Test GetReceivedPower
receivedPower = piSimdis.getReceivedPower(searchRadarId, math.radians(91), 3400, 1200, 82.3, 60, 100, 1300)
if not receivedPower.isOk(2000):
	raise Exception("Failed to get a response rfpropGetReceivedPower")
if not -94.4 <= receivedPower.value() <= -94.2:
	raise Exception('Got Two-way Received Power outside range.  Expected -94.3, got ' + str(receivedPower.value()))

############################################
# Test setPODLossThreshold
msg = RemoteApi_pb2.Command()
threshold = msg.piSimdis.rfprop.setPODLossThreshold;
threshold.beamId = searchRadarId;
for x in range(0, 100):
  threshold.podLoss.append(x/100.0)
f = remoteapi.PluginApi.ResponseHandler(client, client.send(msg), ReturnCodeOnly)
if not f.isOk(2000):
	raise Exception("setPODLossThreshold: " + f.responseDescription())
if f.value() != 0:
	raise Exception("setPODLossThreshold: return value = " + str(f.value()))

############################################
# Test SetColorMap
msg = RemoteApi_pb2.Command()
colorMap = msg.piSimdis.rfprop.setColorMap
colorMap.beamId = searchRadarId
colorMap.mapType = 0
for x in range(0, 100):
  entry = colorMap.threshToColor.add()
  entry.threshold = x/100.0
  entry.color = x
f = remoteapi.PluginApi.ResponseHandler(client, client.send(msg), ReturnCodeOnly)
if not f.isOk(2000):
	raise Exception("setColorMap: " + f.responseDescription())
if f.value() != 0:
	raise Exception("setColorMap: return value = " + str(f.value()))

############################################
# Test SlotExists
msg = RemoteApi_pb2.Command()
slot = msg.piSimdis.rfprop.slotExists
slot.beamId = searchRadarId
slot.trueAzimuth = 92
f = remoteapi.PluginApi.ResponseHandler(client, client.send(msg),
	lambda r: r.piSimdis.rfprop.valid)
if not f.isOk(2000):
	raise Exception("slotExists: " + f.responseDescription())
if f.value() != True:
	raise Exception("slotExists: expecting = True  got = " + str(f.value()))

############################################
# Test valid
msg = RemoteApi_pb2.Command()
msg.piSimdis.rfprop.valid = searchRadarId
f = remoteapi.PluginApi.ResponseHandler(client, client.send(msg),
	lambda r: r.piSimdis.rfprop.valid)
if not f.isOk(2000):
	raise Exception("valid: " + f.responseDescription())
if f.value() != True:
	raise Exception("valid: expecting = True  got = " + str(f.value()))

############################################
# Test GetInputFiles
msg = RemoteApi_pb2.Command()
msg.piSimdis.rfprop.getInputFiles = searchRadarId
f = remoteapi.PluginApi.ResponseHandler(client, client.send(msg),
	lambda r: r.piSimdis.rfprop.filenames)
if not f.isOk(2000):
	raise Exception("getInputFiles: " + f.responseDescription())
if len(f.value()) != 10:
	raise Exception("getInputFiles: expecting = 10 got = " + str(len(f.value())))

############################################
# Test setRF
msg = RemoteApi_pb2.Command()
colorMap = msg.piSimdis.rfprop.setRF
colorMap.beamId = searchRadarId
colorMap.display = True;
colorMap.aglHeightOption = 200
colorMap.aglHeight = 250
colorMap.drawSpace = 1
colorMap.height = 300
colorMap.thickness = 200
colorMap.history = 100
colorMap.transparency = 50
colorMap.thresholdMode = 1
colorMap.thresholdType = 2
colorMap.thresholdValue = 47
colorMap.aboveColor = 0x00ff00ff
colorMap.belowColor = 0xff0000ff
f = remoteapi.PluginApi.ResponseHandler(client, client.send(msg), ReturnCodeOnly)
if not f.isOk(2000):
	raise Exception("setRF: " + f.responseDescription())
if f.value() != 0:
	raise Exception("setRF: return value = " + str(f.value()))

client.close()
print("Test successful")
