import sys
import time
import struct
import zmq
import remoteapi

# How long to sleep after a connect in client
SLEEP_AFTER_CONNECT = 0.1
SLEEP_AFTER_FAILED_SEND = 0.010
SLEEP_AFTER_FAILED_RECEIVE = 0.010

MAX_SEQUENCE_ID = 64*1024 - 1
SEQUENCE_ONE_QUARTER = MAX_SEQUENCE_ID / 4
SEQUENCE_THREE_QUARTERS = SEQUENCE_ONE_QUARTER * 3

# Number of hex digits in client ID
CLIENT_STR_SIZE = 8

# Returns a packed byte array for a big endian 2 byte unsigned integer (unsigned short)
def packUShort_(sequenceId):
	return struct.pack('>H', sequenceId)

# Given a packed byte array of length 2, unpacks a big endian encoded 2 byte unsigned integer (unsigned short) and returns it
def unpackUShort_(sequenceValue):
	return struct.unpack('>H', sequenceValue)[0]

class CommandReplyClient(object):
	# @param zmqContext ZeroMQ context
	def __init__(self, zmqContext, serverAddress, commandPort, responsePort):
		self.valid_ = False
		self.zmqContext_ = zmqContext
		self.serverAddress_ = serverAddress
		self.sequenceNumber_ = 0
		self.expectedRxNumber_ = 0
		self.pendingPayload_ = None
		self.pendingSequenceNumber_ = 0
		self.commandSocket_ = None
		self.responseSocket_ = None
		self.connect_(serverAddress, commandPort, responsePort)
		if self.negotiate_(5000) == remoteapi.CONNECTION_OK:
			self.valid_ = True

	def __del__(self):
		if self.commandSocket_ is not None:
			self.commandSocket_.close()
		if self.responseSocket_ is not None:
			self.responseSocket_.close()

	def isValid(self):
		return self.valid_

	def negotiate_(self, timeoutMs):
		if self.commandSocket_ is None or self.responseSocket_ is None:
			return remoteapi.CONNECTION_INVALID

		# Wait until we can send() successfully
		sendPoller = zmq.Poller()
		sendPoller.register(self.commandSocket_, zmq.POLLOUT)
		socks = dict(sendPoller.poll(timeoutMs))
		if not self.commandSocket_ in socks:
			return remoteapi.CONNECTION_BLOCKED
		self.commandSocket_.send(b'')

		# Wait until we can receive
		recvPoller = zmq.Poller()
		recvPoller.register(self.responseSocket_, zmq.POLLIN)
		socks = dict(recvPoller.poll(timeoutMs))
		if not self.responseSocket_ in socks:
			return remoteapi.CONNECTION_BLOCKED
		data = self.responseSocket_.recv_multipart()
		# We expect to receive back 1 item with no data
		if data is None or len(data) != 1 or len(data[0]) != 0:
			return remoteapi.CONNECTION_BLOCKED
		return remoteapi.CONNECTION_OK

	# Connects to a server
	# @param address Address to connect to, e.g. tcp://localhost
	# @param responsePort Server port for responses
	# @param commandPort Server port for commands
	def connect_(self, address, commandPort, responsePort):
		# Parse the incoming message into parts
		responseAddress = "%s:%d" % (address, responsePort)
		commandAddress = "%s:%d" % (address, commandPort)

		# Initialize our pub and sub connections to the server
		self.responseSocket_ = self.zmqContext_.socket(zmq.PULL)
		self.responseSocket_.setsockopt(zmq.LINGER, 0)
		self.responseSocket_.setsockopt(zmq.RCVHWM, 10000)
		self.responseSocket_.setsockopt(zmq.RCVBUF, 500000)
		self.responseSocket_.connect(responseAddress)
		self.commandSocket_ = self.zmqContext_.socket(zmq.PUSH)
		self.commandSocket_.setsockopt(zmq.LINGER, 0)
		self.commandSocket_.setsockopt(zmq.RCVHWM, 10000)
		self.commandSocket_.setsockopt(zmq.RCVBUF, 500000)
		self.commandSocket_.connect(commandAddress)

	# Sends a message to the publisher.  Returns array:
	# [0]: CONNECTION_ flag indicating the behavior
	# [1]: Sequence number, or None if error
	def sendMessage(self, payload, waitMilliseconds=100):
		if not self.isValid():
			return (remoteapi.CONNECTION_INVALID, None)

		# Form the message to send
		allParts = [packUShort_(self.sequenceNumber_) + payload]
		# Calculate the time of the clock to stop at
		stopAt = 0.001 * waitMilliseconds + time.process_time()
		while True:
			# Attempt to send the multipart message
			try:
				self.commandSocket_.send_multipart(allParts, zmq.NOBLOCK)
				# Send successful, return success and the sent sequence number
				sentSequence = self.sequenceNumber_
				self.sequenceNumber_ = self.sequenceNumber_ + 1
				# Do rollover if needed
				if self.sequenceNumber_ > MAX_SEQUENCE_ID:
					self.sequenceNumber_ = 0
				return (remoteapi.CONNECTION_OK, sentSequence)
			except zmq.ZMQError as e:
				# Send failed, for any reason.  EAGAIN is a valid reason, others are not
				if e.errno != zmq.EAGAIN:
					self.valid_ = False
					return (remoteapi.CONNECTION_INVALID, None)

			# If send failed, sleep on it and try again
			if time.process_time() >= stopAt:
				break
			time.sleep(SLEEP_AFTER_FAILED_SEND)

		# Return that we've blocked on the send
		return (remoteapi.CONNECTION_BLOCKED, None)

	def isSequenceNumberGreater_(self, number1, number2):
		# Make sure that our values are within the size we expect, else this routine needs changes
		if number1 > MAX_SEQUENCE_ID or number2 > MAX_SEQUENCE_ID:
			raise Exception("Invalid numbers, need to update method isSequenceNumberGreater_()")
		# Check for roll-over using quarter marks
		if number1 < SEQUENCE_ONE_QUARTER and number2 > SEQUENCE_THREE_QUARTERS:
			return True
		return number1 > number2

	# Returns Array of:
	# [0]: Connection State (CONNECTION_OK preferred)
	# [1]: Payload if valid, or None
	def getResponse(self, sequenceNumber, waitMilliseconds=1000):
		if not self.isValid():
			return (remoteapi.CONNECTION_INVALID, None)

		# Check for pending responses
		if self.pendingPayload_ is not None:
			# Have not requested up to the currently received response
			if self.isSequenceNumberGreater_(self.pendingSequenceNumber_, sequenceNumber):
				return (remoteapi.CONNECTION_RESPONSE_ALREADY_PASSED, None)
			# Requested the pending response
			if self.pendingSequenceNumber_ == sequenceNumber:
				self.pendingPayload_ = None
				return (remoteapi.CONNECTION_OK, self.pendingPayload_)
			# Went past pending response
			self.pendingPayload_ = None

		stopAt = time.process_time() + 0.001 * waitMilliseconds
		while True:
			response = None
			try:
				response = self.responseSocket_.recv_multipart(zmq.NOBLOCK)
				# Response is a 1-part message -- Sequence ID, Payload
				if len(response) != 1:
					return (remoteapi.CONNECTION_MISSING_MESSAGE_HEADER, None)
			except zmq.ZMQError as e:
				# Send failed, for any reason.  EAGAIN is a valid reason, others are not
				if e.errno != zmq.EAGAIN:
					self.valid_ = False
					return (remoteapi.CONNECTION_INVALID, None)
				# Sleep to prevent busy-locking
				if time.process_time() >= stopAt:
					break
				time.sleep(SLEEP_AFTER_FAILED_RECEIVE)
				continue

			# Make sure the sequence number is what we expect
			gotSequenceId = unpackUShort_(response[0][:2])
			if self.expectedRxNumber_ != gotSequenceId:
				raise Exception("In getResponse, expected: " + str(self.expectedRxNumber_) + " Got: " + str(gotSequenceId) + " Wanted: " + str(sequenceNumber))
			self.expectedRxNumber_ = self.expectedRxNumber_ + 1

			if self.isSequenceNumberGreater_(gotSequenceId, sequenceNumber):
				# Need to save it off and wait for it to be called
				self.pendingSequenceNumber_ = gotSequenceId
				self.pendingPayload_ = response[0][2:]
				return (remoteapi.CONNECTION_RESPONSE_ALREADY_PASSED, None)
			elif gotSequenceId == sequenceNumber:
				return (remoteapi.CONNECTION_OK, response[0][2:])
			# else: drop the response

			# Check for too long waiting, then try again
			if time.process_time() >= stopAt:
				break
			time.sleep(SLEEP_AFTER_FAILED_RECEIVE)
		# If we got here, we ran out of time
		return (remoteapi.CONNECTION_BLOCKED, None)

	# Returns connection state, NOT as an array.  Note that this is like getResponse but doesn't sleep
	def flushResponse(self, sequenceNumber):
		if not self.isValid():
			return remoteapi.CONNECTION_INVALID

		# Check for pending responses
		if self.pendingPayload_ is not None:
			# Have not requested up to the currently received response
			if self.isSequenceNumberGreater_(self.pendingSequenceNumber_, sequenceNumber):
				return remoteapi.CONNECTION_RESPONSE_ALREADY_PASSED
			# Requested the pending response
			if self.pendingSequenceNumber_ == sequenceNumber:
				self.pendingPayload_ = None
				return remoteapi.CONNECTION_OK
			# Went past pending response
			self.pendingPayload_ = None

		while True:
			response = None
			try:
				response = self.responseSocket_.recv_multipart(zmq.NOBLOCK)
				# Response is a 1-part message -- Sequence ID, Payload
				if len(response) != 1:
					return remoteapi.CONNECTION_MISSING_MESSAGE_HEADER
			except zmq.ZMQError as e:
				# Send failed, for any reason.  EAGAIN is a valid reason, others are not
				if e.errno != zmq.EAGAIN:
					self.valid_ = False
					return (remoteapi.CONNECTION_INVALID, None)
				# No data; we didn't clear the ID; so return failure of blockage
				return remoteapi.CONNECTION_BLOCKED

			# Make sure the sequence number is what we expect
			gotSequenceId = unpackUShort_(response[0][:2])
			if self.expectedRxNumber_ != gotSequenceId:
				raise "In getResponse, expected: " + str(self.expectedRxNumber_) + " Got: " + str(gotSequenceId) + " Wanted: " + str(sequenceNumber)
			self.expectedRxNumber_ = self.expectedRxNumber_ + 1

			if self.isSequenceNumberGreater_(gotSequenceId, sequenceNumber):
				# Need to save it off and wait for it to be called
				self.pendingSequenceNumber_ = gotSequenceId
				self.pendingPayload_ = response[0][2:]
				return remoteapi.CONNECTION_RESPONSE_ALREADY_PASSED
			elif gotSequenceId == sequenceNumber:
				return remoteapi.CONNECTION_OK
			# else: drop the response
			pass
		# If we got here, we ran out of time
		return remoteapi.CONNECTION_BLOCKED

	def flushAllResponses(self, waitMilliseconds):
		if not self.isValid():
			return CONNECTION_INVALID
		stopAt = time.process_time() + 0.001 * waitMilliseconds
		while True:
			# Constantly flush the highest sequence number until time runs out
			state = self.flushResponse(MAX_SEQUENCE_ID)
			if state != remoteapi.CONNECTION_OK and state != remoteapi.CONNECTION_BLOCKED and state != remoteapi.CONNECTION_RESPONSE_ALREADY_PASSED:
				return state
			# Depending on sequence numbers, flush may place a message in pending
			pendingPayload_ = None
			# Check for too long waiting, then try again
			if time.process_time() >= stopAt:
				return remoteapi.CONNECTION_OK
			time.sleep(SLEEP_AFTER_FAILED_RECEIVE)
